/*=============================================================================
 * Classe Connection
 * 
 * Gestion de la connexion et lancement des services.
 * WiFi / mDNS / OTA / MQTT
 * 
 * © Jean-Luc Béchennec - 2022
 */

#include "Connection.h"
#include "Config.h"
#include <WiFi.h>
#include <ESPmDNS.h>
#include <PubSubClient.h>

Connection::State Connection::sState = OFFLINE;
uint32_t Connection::sLastDate = 0ul;
uint32_t Connection::sPeriod = 1000ul;
char *Connection::sSsid = NULL;
char *Connection::sPass = NULL;
char *Connection::sName = NULL;
char *Connection::sOTAPass = NULL;
char *Connection::sBrokerName = NULL;
IPAddress Connection::sBrokerIP = {0, 0, 0, 0};
WiFiClient Connection::sNet;
PubSubClient Connection::sClient(Connection::sNet);
Connection::MessageHandlingFunction Connection::sMessageHandler = NULL; 
Connection::SubscriptionFunction Connection::sDoSubscriptions = NULL;


void Connection::update()
{
  switch (sState) {
    case OFFLINE:
      if (sSsid != NULL && sPass != NULL) {
        DP("Connexion à : ");
        WiFi.begin(sSsid, sPass);
        sState = WIFI_STBY;
      }
      break;
      
    case WIFI_STBY:
      if (WiFi.status() == WL_CONNECTED) {
        DP(' ');
        DPLN(sSsid);
        if (sName != NULL) {
          MDNS.begin(sName);
          DP("Demarrage de mDNS sous le nom ");
          DP(sName);
          DPLN(".local");
        } else {
          DPLN("Pas de nom mDNS");
        }
        sState = WIFI_OK;
      } else {
        DP('.');
      }
      break;
      
    case WIFI_OK:
      if (WiFi.status() == WL_CONNECTED) {
        if (sBrokerName != NULL) {
          DP("Recherche du Broker ");
          DP(sBrokerName);
          DP(".local - ");
          sBrokerIP = MDNS.queryHost(sBrokerName);
          if (sBrokerIP != IPAddress(0, 0, 0, 0)) {
            DPLN("trouvé !");
            sState = MDNS_OK;
          } else {
            DPLN("absent");
          }
        }
      } else {
        DPLN("Perte de connexion");
        sState = OFFLINE;
      }
      break;

    case MDNS_OK:
      if (WiFi.status() == WL_CONNECTED) {
        initOTA();
        sClient.setServer(sBrokerIP, 1883);
        sClient.setCallback(incomingMessage);
        sClient.setKeepAlive(kMQTTBrokerKeepAlive);
        sClient.setSocketTimeout(kMQTTSocketTimeout);
        sState = OTA_OK;
      } else {
        DPLN("Perte de connexion");
        sState = OFFLINE;
      }
      break;

    case OTA_OK:
      if (WiFi.status() == WL_CONNECTED) {
        DP("Connexion au Broker ");
        DP(sBrokerName);
        DP(", IP = ");
        DP(sBrokerIP);
        DP(" - ");
        if (sClient.connect(sName)) {
          DPLN("Ok");
          performSubscriptions();
          sState = MQTT_OK;
        }
        else {
          DPLN("Refusée"); 
        }
      } else {
        DPLN("Perte de connexion");
        ArduinoOTA.end();
        sState = OFFLINE;
      }
      break;

    case MQTT_OK:
      if (WiFi.status() == WL_CONNECTED) {
        if (! sClient.connected()) {
          DPLN("Perte de connexion au Broker");  
          sState = OTA_OK;    
        }
      } else {
        DPLN("Perte de connexion");
        ArduinoOTA.end();
        sState = OFFLINE;
      }

  }
}

void Connection::initOTA()
{
  DP("Demarrage OTA sous le nom ");
  DPLN(sName);
  ArduinoOTA.setHostname(sName);
  if (sOTAPass != NULL) {
    ArduinoOTA.setPassword(sOTAPass);
  }
  
#ifdef DEBUG
  ArduinoOTA.onStart(startingOTA);
  ArduinoOTA.onProgress(OTAInProgress);
  ArduinoOTA.onEnd(endingOTA);
  ArduinoOTA.onError(OTAError);
#endif

  ArduinoOTA.begin();
}

void Connection::incomingMessage(char *inTopic, byte *inPayload, unsigned int inLength)
{
  if (sMessageHandler != NULL) {
    sMessageHandler(inTopic, inPayload, inLength);
  }
}

void Connection::performSubscriptions()
{
  if (sDoSubscriptions != NULL) {
    sDoSubscriptions();
  }
}

void Connection::attach(SubscriptionFunction inSubFunction, MessageHandlingFunction inMessageHandler)
{
  sDoSubscriptions = inSubFunction;
  sMessageHandler = inMessageHandler;
}


#ifdef DEBUG
void Connection::startingOTA()
{
  const int command = ArduinoOTA.getCommand();
  if (command == U_FLASH) {
    DPLN("Téléversement du firmware");
  } else {
    DP("Commande inattendue ");
    DPLN(command);
  }
}

void Connection::OTAInProgress(unsigned int inProgress, unsigned int inTotal)
{
  static uint32_t previousNumberOfDots = 0;
  const uint32_t numberOfDots = 50 * inProgress / inTotal;
  for (uint32_t dot = previousNumberOfDots; dot < numberOfDots; dot++) {
    DP('.');
  }
  if (inTotal == inProgress) {
    DPLN();
  }
  previousNumberOfDots = numberOfDots;
}

void Connection::endingOTA()
{
  DPLN("Téléversement terminé");
}

void Connection::OTAError(ota_error_t error)
{
  DP("Erreur[");
  DP(error);
  DP("] : ");
  switch (error) {
    case OTA_AUTH_ERROR:    DPLN("L'authentification a échoué"); break;
    case OTA_BEGIN_ERROR:   DPLN("Échec au début");              break;
    case OTA_CONNECT_ERROR: DPLN("Échec à la connexion");        break;
    case OTA_RECEIVE_ERROR: DPLN("Échec à la réception");        break;
    case OTA_END_ERROR:     DPLN("Échec à la fermeture");        break;
  }
}
#endif


void Connection::setup(const char *inSsid,
                       const char *inPass,
                       const char *inName,
                       const char *inBrokerName,
                       const char *inOTAPass,
                       const uint32_t inPeriod)
{
  sPeriod = inPeriod;
  sSsid = (char *)inSsid;
  sPass = (char *)inPass;
  sName = (char *)inName;
  sBrokerName = (char *)inBrokerName;
  sOTAPass = (char *)inOTAPass;
}

void Connection::loop()
{
  const uint32_t currentDate = millis();
  if ((currentDate - sLastDate) >= sPeriod) {
    update();
    sLastDate = currentDate;
  }
  ArduinoOTA.handle();
  sClient.loop();
}

bool Connection::isOnline()
{ 
  return (sState == MQTT_OK);
}

void Connection::subscribe(const char *inTopic)
{
  sClient.subscribe(inTopic);
}

void Connection::publish(const char *inTopic, const float inValue)
{
  String s(inValue);
  sClient.publish(inTopic, s.c_str());
}

void Connection::publish(const char *inTopic, const double inValue)
{
  String s(inValue);
  sClient.publish(inTopic, s.c_str());
}

void Connection::publish(const char *inTopic, const uint32_t inValue)
{
  String s(inValue);
  sClient.publish(inTopic, s.c_str());
}

void Connection::publish(const char *inTopic, const int32_t inValue)
{
  String s(inValue);
  sClient.publish(inTopic, s.c_str());
}

void Connection::publish(const char *inTopic, const char *inValue)
{
  sClient.publish(inTopic, inValue);
}

void Connection::publish(const char *inTopic, String &inValue)
{
  sClient.publish(inTopic, inValue.c_str());
}
